chiark / gitweb /
[PATCH] Revert previous dev_d.c change, it's not what is causing HAL problems.
[elogind.git] / namedev.c
index 7f3a6a7783913c6bd89b49d360013b72e4176f5a..695fb55413910e7507b408425529875f1db4c759 100644 (file)
--- a/namedev.c
+++ b/namedev.c
@@ -42,6 +42,7 @@
 #include "logging.h"
 #include "namedev.h"
 #include "klibc_fixups.h"
+#include "udevdb.h"
 
 static struct sysfs_attribute *find_sysfs_attribute(struct sysfs_class_device *class_dev, struct sysfs_device *sysfs_device, char *attr);
 
@@ -179,6 +180,37 @@ static int get_format_len(char **str)
        return -1;
 }
 
+/** Finds the lowest positive N such that <name>N isn't present in 
+ *  $(udevroot) either as a file or a symlink.
+ *
+ *  @param  name                Name to check for
+ *  @return                     0 if <name> didn't exist and N otherwise.
+ */
+static unsigned int find_free_number (struct udevice *udev, char *name)
+{
+       char temp[NAME_SIZE];
+       char path[NAME_SIZE];
+       struct udevice dev;
+       int result;
+
+       /* have to sweep the database for each lookup */
+       result = 0;
+       strncpy(temp, name, sizeof (temp));
+       while (1) {
+               if (udevdb_get_dev_byname(temp, path, &dev) != 0)
+                       goto found;
+               /* symlink might be stale if $(udevroot) isn't cleaned; check
+                * on major/minor to see if it's the same device
+                */
+               if (dev.major == udev->major && dev.minor == udev->minor)
+                       goto found;
+               snprintf (temp, sizeof(temp), "%s%d", name, ++result);
+       }
+
+found:
+       return result;
+}
+
 static void apply_format(struct udevice *udev, char *string, size_t maxsize,
                         struct sysfs_class_device *class_dev,
                         struct sysfs_device *sysfs_device)
@@ -195,6 +227,7 @@ static void apply_format(struct udevice *udev, char *string, size_t maxsize,
        char *rest;
        int slen;
        struct sysfs_attribute *tmpattr;
+       unsigned int next_free_number;
 
        pos = string;
        while (1) {
@@ -284,6 +317,13 @@ static void apply_format(struct udevice *udev, char *string, size_t maxsize,
                        strfieldcatmax(string, "%", maxsize);
                        pos++;
                        break;
+               case 'e':
+                       next_free_number = find_free_number(udev, string);
+                       if (next_free_number > 0) {
+                               snprintf(temp2, sizeof(temp2), "%d", next_free_number);
+                               strfieldcatmax(string, temp2, maxsize);
+                       }
+                       break;
                default:
                        dbg("unknown substitution type '%%%c'", c);
                        break;
@@ -372,7 +412,7 @@ static void fix_kernel_name(struct udevice *udev)
        }
 }
 
-static int execute_program(char *path, char *value, int len)
+static int execute_program(const char *path, char *value, int len)
 {
        int retval;
        int count;
@@ -381,28 +421,34 @@ static int execute_program(char *path, char *value, int len)
        pid_t pid;
        char *pos;
        char arg[PROGRAM_SIZE];
-       char *argv[sizeof(arg) / 2];
+       char *argv[(PROGRAM_SIZE / 2) + 1];
        int i;
 
+       strfieldcpy(arg, path);
        i = 0;
        if (strchr(path, ' ')) {
-               strfieldcpy(arg, path);
                pos = arg;
                while (pos != NULL) {
                        if (pos[0] == '\'') {
                                /* don't separate if in apostrophes */
                                pos++;
                                argv[i] = strsep(&pos, "\'");
-                               while (pos[0] == ' ')
+                               while (pos && pos[0] == ' ')
                                        pos++;
-               } else {
+                       } else {
                                argv[i] = strsep(&pos, " ");
                        }
                        dbg("arg[%i] '%s'", i, argv[i]);
                        i++;
                }
+               argv[i] =  NULL;
+               dbg("execute '%s' with parsed arguments", arg);
+       } else {
+               argv[0] = arg;
+               argv[1] = main_argv[1];
+               argv[2] = NULL;
+               dbg("execute '%s' with subsystem '%s' argument", arg, argv[1]);
        }
-       argv[i] =  NULL;
 
        retval = pipe(fds);
        if (retval != 0) {
@@ -414,17 +460,9 @@ static int execute_program(char *path, char *value, int len)
        switch(pid) {
        case 0:
                /* child */
-               close(STDOUT_FILENO);
-
-               /* dup write side of pipe to STDOUT */
-               dup(fds[1]);
-               if (argv[0] !=  NULL) {
-                       dbg("execute '%s' with given arguments", argv[0]);
-                       retval = execv(argv[0], argv);
-               } else {
-                       dbg("execute '%s' with main argument", path);
-                       retval = execv(path, main_argv);
-               }
+               /* dup2 write side of pipe to STDOUT */
+               dup2(fds[1], STDOUT_FILENO);
+               retval = execv(arg, argv);
 
                info(FIELD_PROGRAM " execution of '%s' failed", path);
                exit(1);
@@ -604,9 +642,8 @@ static int match_place(struct config_device *dev, struct sysfs_class_device *cla
 
 static int whitelist_search(struct sysfs_class_device *class_dev)
 {
+       char *sysblock = "/sys/block";
        int i;
-       int list_size = 17;
-       int sysblock = 10;
 
        static char *list[] = {
                "nb",
@@ -626,15 +663,15 @@ static int whitelist_search(struct sysfs_class_device *class_dev)
                "pf",
                "scd",
                "ubd",
+               NULL,
        };
 
-       if (!strncmp(class_dev->path, "/sys/block", sysblock)) {
-               for (i=0; i < list_size; i++) {
-                       if (!strncmp(class_dev->name, list[i], strlen(list[i]))) {
-                               return 1;
-                       }
-               }
-       }
+       if (strncmp(class_dev->path, sysblock, strlen(sysblock)))
+               return 0;
+
+       for (i=0; list[i] != NULL; i++)
+               if (!strncmp(class_dev->name, list[i], strlen(list[i])))
+                       return 1;
 
        return 0;
 }
@@ -774,10 +811,12 @@ static int match_rule(struct config_device *dev, struct sysfs_class_device *clas
 
                /* execute external program */
                if (dev->program[0] != '\0') {
+                       char program[PROGRAM_SIZE];
+
                        dbg("check " FIELD_PROGRAM);
-                       apply_format(udev, dev->program, sizeof(dev->program),
-                                    class_dev, sysfs_device);
-                       if (execute_program(dev->program, udev->program_result, NAME_SIZE) != 0) {
+                       strfieldcpy(program, dev->program);
+                       apply_format(udev, program, sizeof(program), class_dev, sysfs_device);
+                       if (execute_program(program, udev->program_result, NAME_SIZE) != 0) {
                                dbg(FIELD_PROGRAM " returned nonzero");
                                goto try_parent;
                        } else {